#! /usr/bin/env python3

# This script uses the jq255 implementation (jq255.py) to generate test
# vectors. One command-line argument:
#    jq255e    make test vectors for jq255e
#    jq255s    make test vectors for jq255s
# If no argument is provided, then 'jq255e' is assumed.
#
# Test vectors are produced on standard output. Test vector production
# is deterministic (you always get the same vectors for a given curve).

import importlib
import sys
import hashlib
import json
import jq255

def add_test(tl, testtype, vv):
    vv['id'] = '%s%d' % (testtype, len(tl))
    tl.append(vv)

# Make test vectors for decoding tests.
def mktests_decode(curve):
    tl = []

    # The all-zero input is decodable.
    add_test(tl, 'decode', { "data": "0000000000000000000000000000000000000000000000000000000000000000", "decodable": True })

    # Make some pseudorandom decodable inputs.
    rs = hashlib.sha256(curve.bname + b'-test-decode').digest()
    ap = -2*curve.a
    bp = curve.a**2 - 4*curve.b
    for i in range(0, 20):
        while True:
            rs = hashlib.sha256(rs).digest()
            rt = int.from_bytes(rs, byteorder='big')
            u = curve.K(rt)
            if (bp*u**4 + ap*u**2 + 1).is_square():
                break
        val = bytes(u)
        # Try decoding, it should work.
        P = curve.Decode(val)
        val2 = bytes(P)
        if val != val2:
            raise Exception('Decode/encode failed (different bytes)')
        add_test(tl, 'decode', { "data": val.hex(), "decodable": True })

    # Make some non-decodable inputs because u is out of range.
    for i in range(0, 20):
        if i == 0:
            # First test checks that values just above the modulus are
            # properly rejected, instead of being implicitly reduced.
            iu = curve.K.m + 1
            while True:
                u = curve.K(iu)
                if (bp*u**4 + ap*u**2 + 1).is_square():
                    break
                iu += 1
        else:
            # Even-numbered tests verify that the top bit is checked but
            # not ignored; odd-numbered tests verify that the whole value
            # is not implicitly reduced.
            while True:
                rs = hashlib.sha256(rs).digest()
                iu = int.from_bytes(rs, byteorder='big')
                if (i & 1) == 0:
                    if iu < 2**255:
                        u = curve.K(iu)
                        iu += 2**255
                    else:
                        u = curve.K(iu - 2**255)
                else:
                    if iu < 2**255:
                        iu += 2**255
                    u = curve.K(iu)
                if (bp*u**4 + ap*u**2 + 1).is_square():
                    break
        val = iu.to_bytes(32, byteorder='little')
        # Try decoding, it should fail.
        good = False
        try:
            curve.Decode(val)
        except Exception:
            good = True
        if not good:
            raise Exception('Decoding should have failed')
        add_test(tl, 'decode', { "data": val.hex(), "decodable": False })

    # Make some non-decodable inputs (u is in range, but matches no point).
    for i in range(0, 20):
        while True:
            rs = hashlib.sha256(rs).digest()
            rt = int.from_bytes(rs, byteorder='big')
            u = curve.K(rt)
            if not((bp*u**4 + ap*u**2 + 1).is_square()):
                break
        val = bytes(u)
        # Try decoding, it should fail.
        good = False
        try:
            curve.Decode(val)
        except Exception:
            good = True
        if not good:
            raise Exception('Decoding should have failed')
        add_test(tl, 'decode', { "data": val.hex(), "decodable": False })

    return tl

# Make test vectors for map-to-curve tests.
def mktests_map(curve):
    tl = []
    rs = hashlib.sha256(curve.bname + b'-test-map').digest()
    for i in range(0, 40):
        rs = hashlib.sha256(rs).digest()
        if i == 0:
            bb = bytearray(32)
            if not curve.a.is_zero():
                bb[0] = 1
        else:
            bb = rs
        P = curve.MapToCurve(bb)
        if i == 0:
            if not P.is_neutral():
                raise Exception('zero should be mapped to neutral')
        else:
            # Sanity checks on the point.
            if P.Z.is_zero():
                raise Exception('invalid point')
            e = P.E / P.Z
            u = P.U / P.Z
            if P.U**2 != P.T*P.Z:
                raise Exception('invalid point')
            if e**2 != curve.bp*u**4 + curve.ap*u**2 + curve.K.one:
                raise Exception('invalid point')
            x = (e + curve.K.one - curve.a*u**2) / (2*u**2)
            w = curve.K.one / u
            if w**2*x != x**2 + curve.a*x + curve.b:
                raise Exception('invalid point')
            if x.is_square():
                raise Exception('mapped to an r-torsion point')
        add_test(tl, 'map', { "f": bb.hex(), "p": bytes(P).hex() })

    return tl

# Make test vectors for point addition.
def mktests_add(curve):
    tl = []
    rs = hashlib.sha256(curve.bname + b'-test-add').digest()
    for i in range(0, 20):
        rs = hashlib.sha256(rs).digest()
        rt = int.from_bytes(rs, byteorder='big')
        P1 = curve.G * rt
        rs = hashlib.sha256(rs).digest()
        rt = int.from_bytes(rs, byteorder='big')
        P2 = curve.G * rt
        P3 = P1 + P2
        P4 = P1 + P1
        P5 = P4 + P2
        P6 = P3 + P3
        add_test(tl, 'add', { "p1": bytes(P1).hex(), "p2": bytes(P2).hex(), "p3": bytes(P3).hex(), "p4": bytes(P4).hex(), "p5": bytes(P5).hex(), "p6": bytes(P6).hex() })

    return tl

# Make test vectors for point multiplication.
# Scalars may range up to the full 256-bit range.
def mktests_mul(curve):
    tl = []
    rs = hashlib.sha256(curve.bname + b'-test-pointmul').digest()
    for i in range(0, 20):
        rs = hashlib.sha256(rs).digest()
        rt = int.from_bytes(rs, byteorder='big')
        P1 = curve.G * rt
        rs = hashlib.sha256(rs).digest()
        rt = curve.SF(int.from_bytes(rs, byteorder='big'))
        P3 = P1 * rt
        add_test(tl, 'mul', { "p": bytes(P1).hex(), "q": bytes(P3).hex(), "s": bytes(rt).hex() })

    return tl

def mktests_keygen(curve):
    tl = []
    for i in range(0, 20):
        sh = hashlib.shake_256()
        sh.update(int(i).to_bytes(1, 'little'))
        sk = jq255.Keygen(curve, sh)
        pk = jq255.MakePublic(curve, sk)
        add_test(tl, 'keygen', { "sk": bytes(sk).hex(), "pk": bytes(pk).hex() })
    return tl

def mktests_hashraw(curve):
    tl = []
    data = bytearray()
    for i in range(0, 100):
        P = jq255.HashToCurve(curve, b'', data)
        add_test(tl, 'hashraw', { "raw": data.hex(), "p": bytes(P).hex() })
        data.append(i)
    return tl

def mktests_hashph(curve):
    tl = []
    data = bytearray()
    for i in range(0, 10):
        sh = hashlib.blake2s()
        sh.update(i.to_bytes(1, byteorder='little'))
        data = sh.digest()
        P = jq255.HashToCurve(curve, jq255.HASHNAME_BLAKE2S, data)
        add_test(tl, 'hashph', { "hv": data.hex(), "p": bytes(P).hex() })
    return tl

def mktests_ECDH(curve):
    tl = []
    for i in range(0, 20):
        # Valid test.
        rng = hashlib.shake_256()
        rng.update(curve.bname)
        rng.update(b'-test-ECDH-self-')
        rng.update(int(i).to_bytes(2, 'little'))
        sk_self = jq255.Keygen(curve, rng)
        pk_self = jq255.MakePublic(curve, sk_self)
        rng = hashlib.shake_256()
        rng.update(curve.bname)
        rng.update(b'-test-ECDH-peer-')
        rng.update(int(i).to_bytes(2, 'little'))
        sk_peer = jq255.Keygen(curve, rng)
        pk_peer = jq255.MakePublic(curve, sk_peer)
        secret, ok = jq255.ECDH(sk_self, pk_self, pk_peer)
        if not ok:
            raise Exception('ECDH failed')
        # Try again with decoding inside the function.
        secret2, ok2 = jq255.ECDH(sk_self, pk_self, bytes(pk_peer))
        if not ok2:
            raise Exception('ECDH failed (with decoding)')
        if secret2 != secret:
            raise Exception('ECDH wrong secret (with decoding)')
        # Verify that the peer would get the same value.
        secret3, ok3 = jq255.ECDH(sk_peer, pk_peer, pk_self)
        if not ok3:
            raise Exception('ECDH failed (reverse)')
        if secret3 != secret:
            raise Exception('ECDH wrong secret (reverse)')

        # Invalid test.
        j = 0
        while True:
            rng = hashlib.shake_256()
            rng.update(curve.bname)
            rng.update(b'-test-ECDH-peer-')
            rng.update(int(i).to_bytes(2, 'little'))
            rng.update(int(j).to_bytes(4, 'little'))
            u = curve.K.DecodeReduce(rng.digest(32))
            if not((curve.bp*u**4 + curve.ap*u**2 + curve.K.one).is_square()):
                break
            j += 1
        # We use the binary encoding of the peer public key, so that the
        # ECDH() function applies the alternate secret generation in case
        # of decoding failure.
        secretbad, ok = jq255.ECDH(sk_self, pk_self, bytes(u))
        if ok:
            raise Exception('ECDH should have failed')

        add_test(tl, 'ECDH', { "sk": bytes(sk_self).hex(), "pk1": bytes(pk_peer).hex(), "sec1": secret.hex(), "pk2": bytes(u).hex(), "sec2": secretbad.hex() })

    return tl

def mktests_sign(curve):
    tl = []
    for i in range(0, 20):
        rng = hashlib.shake_256()
        rng.update(curve.bname)
        rng.update(b'-test-sign-sk-')
        rng.update(i.to_bytes(2, 'little'))
        sk = jq255.Keygen(curve, rng)
        pk = jq255.MakePublic(curve, sk)
        rng.update(curve.bname)
        rng.update(b'-test-sign-seed-')
        rng.update(i.to_bytes(2, 'little'))
        seed = rng.digest(i)
        msg = 'sample {0}'.format(i).encode('utf-8')
        h = hashlib.blake2s()
        h.update(msg)
        hv = h.digest()
        sig = jq255.Sign(sk, pk, jq255.HASHNAME_BLAKE2S, hv, seed)
        if not(jq255.Verify(pk, sig, jq255.HASHNAME_BLAKE2S, hv)):
            raise Exception("Cannot verify signature!")
        if jq255.Verify(pk, sig, jq255.HASHNAME_SHA3_256, hv + b'.'):
            raise Exception("Signature verification should have failed!")
        add_test(tl, 'sign', { "sk": bytes(sk).hex(), "pk": bytes(pk).hex(), "seed": seed.hex(), "msg": msg.hex(), "hv": hv.hex(), "sig": sig.hex() })

    return tl

def mktests(curve):
    tm = {}
    tm["decode"] = mktests_decode(curve)
    tm["map"] = mktests_map(curve)
    tm["add"] = mktests_add(curve)
    tm["mul"] = mktests_mul(curve)
    tm["keygen"] = mktests_keygen(curve)
    tm["hashraw"] = mktests_hashraw(curve)
    tm["hashph"] = mktests_hashph(curve)
    tm["ECDH"] = mktests_ECDH(curve)
    tm["sign"] = mktests_sign(curve)
    return tm

if len(sys.argv) >= 2:
    name = sys.argv[1].lower()
    if name == 'jq255e':
        curve = jq255.Jq255e
    elif name == 'jq255s':
        curve = jq255.Jq255s
    else:
        raise Exception('unknown curve name: %s' % name)
else:
    curve = jq255.Jq255e
print(json.dumps(mktests(curve), indent=2))



#! /usr/bin/env python3

import importlib
import sys
import json
import jq255

def check_decode(curve, tl):
    print('  decode:  ', flush=True, end='')
    for t in tl:
        data = bytes.fromhex(t['data'])
        decodable = t['decodable']
        if decodable:
            p = curve.Decode(data)
            bb = bytes(p)
            if bb != data:
                raise Exception('bad decode/encode')
        else:
            good = True
            try:
                curve.Decode(data)
            except Exception:
                good = False
            if good:
                raise Exception('decoding should have failed')
    print('OK', flush=True)

def check_map(curve, tl):
    print('  map:     ', flush=True, end='')
    for t in tl:
        q = curve.MapToCurve(bytes.fromhex(t['f']))
        if bytes(q) != bytes.fromhex(t['p']):
            raise Exception('bad map-to-group')
    print('OK', flush=True)

def check_add(curve, tl):
    print('  add:     ', flush=True, end='')
    for t in tl:
        ep1 = bytes.fromhex(t['p1'])
        ep2 = bytes.fromhex(t['p2'])
        ep3 = bytes.fromhex(t['p3'])
        ep4 = bytes.fromhex(t['p4'])
        ep5 = bytes.fromhex(t['p5'])
        ep6 = bytes.fromhex(t['p6'])
        p1 = curve.Decode(ep1)
        p2 = curve.Decode(ep2)
        p3 = p1 + p2
        if bytes(p3) != ep3:
            raise Exception('add 1')
        p4 = p1.Double()
        if bytes(p4) != ep4:
            raise Exception('add 2')
        q4 = p1 + p1
        if bytes(q4) != ep4:
            raise Exception('add 3')
        p5 = p4 + p2
        if bytes(p5) != ep5:
            raise Exception('add 4')
        q5 = p1 + p3
        if bytes(q5) != ep5:
            raise Exception('add 5')
        p6 = p3.Double()
        if bytes(p6) != ep6:
            raise Exception('add 6')
        q6 = p4 + p2.Double()
        if bytes(q6) != ep6:
            raise Exception('add 7')
        r3 = q6 - p3
        if bytes(r3) != ep3:
            raise Exception('add 8')
        r2 = q6 - q5
        if bytes(r2) != ep2:
            raise Exception('add 9')
        q = p1
        for j in range(1, 10):
            q += q
            if bytes(q) != bytes(p1.Xdouble(j)):
                raise Exception('xdouble %d' % j)
    print('OK', flush=True)

def check_mul(curve, tl):
    print('  mul:     ', flush=True, end='')
    for t in tl:
        p = curve.Decode(bytes.fromhex(t['p']))
        s = curve.SF.Decode(bytes.fromhex(t['s']))
        if bytes(s*p) != bytes.fromhex(t['q']):
            raise Exception('mul')
    print('OK', flush=True)

def check_keygen(curve, tl):
    print('  keygen:  ', flush=True, end='')
    for t in tl:
        sk = jq255.DecodePrivate(curve, bytes.fromhex(t['sk']))
        pk = jq255.MakePublic(curve, sk)
        if jq255.EncodePublic(pk) != bytes.fromhex(t['pk']):
            raise Exception('keygen')
    print('OK', flush=True)

def check_hashraw(curve, tl):
    print('  hashraw: ', flush=True, end='')
    for t in tl:
        raw = bytes.fromhex(t['raw'])
        p = jq255.HashToCurve(curve, None, raw)
        if bytes(p) != bytes.fromhex(t['p']):
            raise Exception('hashraw')
    print('OK', flush=True)

def check_hashph(curve, tl):
    print('  hashph:  ', flush=True, end='')
    for t in tl:
        hv = bytes.fromhex(t['hv'])
        p = jq255.HashToCurve(curve, jq255.HASHNAME_BLAKE2S, hv)
        if bytes(p) != bytes.fromhex(t['p']):
            raise Exception('hashph')
    print('OK', flush=True)

def check_ECDH(curve, tl):
    print('  ECDH:    ', flush=True, end='')
    for t in tl:
        sk = jq255.DecodePrivate(curve, bytes.fromhex(t['sk']))
        q = jq255.MakePublic(curve, sk)
        pk1 = bytes.fromhex(t['pk1'])
        (sec1, ok1) = jq255.ECDH(sk, q, pk1)
        if not(ok1) or sec1 != bytes.fromhex(t['sec1']):
            raise Exception('ECDH 1')
        pk2 = bytes.fromhex(t['pk2'])
        (sec2, ok2) = jq255.ECDH(sk, q, pk2)
        if ok2 or sec2 != bytes.fromhex(t['sec2']):
            raise Exception('ECDH 2')
    print('OK', flush=True)

def check_sign(curve, tl):
    print('  sign:    ', flush=True, end='')
    for t in tl:
        sk = jq255.DecodePrivate(curve, bytes.fromhex(t['sk']))
        pk = jq255.MakePublic(curve, sk)
        if jq255.EncodePublic(pk) != bytes.fromhex(t['pk']):
            raise Exception('sign 1')
        seed = bytes.fromhex(t['seed'])
        hv = bytes.fromhex(t['hv'])
        sig = jq255.Sign(sk, pk, jq255.HASHNAME_BLAKE2S, hv, seed)
        if sig != bytes.fromhex(t['sig']):
            raise Exception('sign 2')
        sig = bytearray(sig)
        hv = bytearray(hv)
        if not(jq255.Verify(pk, sig, jq255.HASHNAME_BLAKE2S, hv)):
            raise Exception('sign 3')
        hv[11] ^= 0x01
        if jq255.Verify(pk, sig, jq255.HASHNAME_BLAKE2S, hv):
            raise Exception('sign 4')
        hv[11] ^= 0x01
        if not(jq255.Verify(pk, sig, jq255.HASHNAME_BLAKE2S, hv)):
            raise Exception('sign 5')
        sig[13] ^= 0x01
        if jq255.Verify(pk, sig, jq255.HASHNAME_BLAKE2S, hv):
            raise Exception('sign 6')
    print('OK', flush=True)

def check(curve):
    f = open('test-%s.json' % curve.name)
    tm = json.load(f)
    f.close()
    print('Tests %s:' % curve.name, flush=True)
    check_decode(curve, tm['decode'])
    check_map(curve, tm['map'])
    check_add(curve, tm['add'])
    check_mul(curve, tm['mul'])
    check_keygen(curve, tm['keygen'])
    check_hashraw(curve, tm['hashraw'])
    check_hashph(curve, tm['hashph'])
    check_ECDH(curve, tm['ECDH'])
    check_sign(curve, tm['sign'])

check(jq255.Jq255e)
check(jq255.Jq255s)



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "sympy",
#     "mpmath",
# ]
# ///

from sympy import isprime, primerange
from random import randrange, shuffle
from mpmath import li
from sys import argv

BIT_SIZE = int(argv[1]) // 2 if len(argv) > 1 else 1024

PRIMES: list[int] = list(primerange(3, 1000000)) # type: ignore

# Ratio of odd integers less than 2^BIT_SIZE that are prime.
# Li(2^BIT_SIZE) / 2^BIT_SIZE * 2
PRIME_RATIO = float(li(2**BIT_SIZE) / 2**(BIT_SIZE - 1))

# Ratio of odd integers divisible by primes[:n].
SMALLDIV_RATIOS: list[float] = [0]

for p in PRIMES:
    SMALLDIV_RATIOS.append(1 - (1 - SMALLDIV_RATIOS[-1]) * ((p - 1) / p))

count = int(round(1 / PRIME_RATIO, 0))

# Buckets of values in each expected category.
small_divisor = []
composite = {"want": count - int(round(count * SMALLDIV_RATIOS[-1])) - 1, "have": []}
prime = {"want": 1, "have": []}

prev = 0
for i in range(1, len(SMALLDIV_RATIOS)):
    want = int(round(count * SMALLDIV_RATIOS[i] - prev, 0))
    if want > 0:
        prev += want
        small_divisor.append({
            "nprimes": i,
            "want": want,
            "have": [],
        })

done = lambda: all(len(bucket["have"]) >= bucket["want"] for bucket in small_divisor) and \
    len(composite["have"]) >= composite["want"] and len(prime["have"]) >= prime["want"]

while not done():
    x = randrange(2**(BIT_SIZE-1) + 2**(BIT_SIZE-2) + 1, 2**BIT_SIZE, 2) | 0b111
    if isprime(x):
        prime["have"].append(x)
        continue
    for n, p in enumerate(PRIMES, start=1):
        if x % p == 0:
            for bucket in small_divisor:
                if n <= bucket["nprimes"]:
                    bucket["have"].append(x)
                    break
            break
    else:
        composite["have"].append(x)

out = []
for bucket in small_divisor:
    out.extend(bucket["have"][:bucket["want"]])
out.extend(composite["have"][:composite["want"]])
shuffle(out)
out.extend(prime["have"][:prime["want"]])

for x in out:
    print(hex(x)[2:])
